/********************************************************************************
 * Copyright (c) 2020-2023 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/
import {
    boundsFeature,
    connectableFeature,
    Containable,
    containerFeature,
    deletableFeature,
    EditableLabel,
    fadeFeature,
    GChildElement,
    GCompartment,
    GGraph,
    GModelElement,
    GModelElementSchema,
    GShapeElement,
    hoverFeedbackFeature,
    isEditableLabel,
    isGModelElementSchema,
    layoutableChildFeature,
    layoutContainerFeature,
    moveFeature,
    Nameable,
    nameFeature,
    openFeature,
    popupFeature,
    RectangularNode,
    resizeFeature,
    selectFeature,
    WithEditableLabel,
    withEditLabelFeature
} from '@eclipse-glsp/client';
import { PNMLModelTypes } from './model-types';

export class TransitionNode extends RectangularNode implements Nameable, WithEditableLabel {
    static override readonly DEFAULT_FEATURES = [
        connectableFeature,
        deletableFeature,
        selectFeature,
        boundsFeature,
        moveFeature,
        layoutContainerFeature,
        fadeFeature,
        hoverFeedbackFeature,
        popupFeature,
        nameFeature,
        withEditLabelFeature,
        resizeFeature
    ];
    duration?: number;
    taskType?: string;
    reference?: string;

    get editableLabel(): (GChildElement & EditableLabel) | undefined {
        const label = this.children.find(element => element.type === 'label:heading');
        if (label && isEditableLabel(label)) {
            return label;
        }
        return undefined;
    }

    get name(): string {
        const labelText = this.editableLabel?.text;
        return labelText ? labelText : '<unknown>';
    }
}

export class InsertableGCompartment extends GCompartment implements Containable {
    static override readonly DEFAULT_FEATURES = [
        boundsFeature,
        layoutContainerFeature,
        layoutableChildFeature,
        fadeFeature,
        containerFeature
    ];

    isContainableElement(_input: string | GShapeElement | GModelElementSchema): boolean {
        return true;
    }
}

export class CategoryNode extends RectangularNode implements Nameable, WithEditableLabel, Containable {
    static override readonly DEFAULT_FEATURES = [
        deletableFeature,
        selectFeature,
        boundsFeature,
        moveFeature,
        layoutContainerFeature,
        fadeFeature,
        hoverFeedbackFeature,
        popupFeature,
        nameFeature,
        withEditLabelFeature,
        openFeature,
        containerFeature
    ];

    name = '';

    isContainableElement(input: GModelElement | GModelElementSchema | string): boolean {
        return true;
    }

    get editableLabel(): (GChildElement & EditableLabel) | undefined {
        const label = this.children.find(element => element.type === 'label:heading');
        if (label && isEditableLabel(label)) {
            return label;
        }
        return undefined;
    }
}

export function getTypeString(input: GModelElement | GModelElementSchema | string): string {
    if (input instanceof GModelElement || isGModelElementSchema(input)) {
        return input.type;
    } else {
        return input;
    }
}

export class PageNode extends CategoryNode {
    override isContainableElement(input: GModelElement | GModelElementSchema | string): boolean {
        const type: string = getTypeString(input);
        if (type === PNMLModelTypes.SERVICE_CHANNEL) {
            return false;
        }
        return true;
    }
}

export class ServerPageNode extends CategoryNode {
    override isContainableElement(input: GModelElement | GModelElementSchema | string): boolean {
        const type: string = getTypeString(input);
        if (
            type === PNMLModelTypes.PLACE ||
            type === PNMLModelTypes.TRANSITION ||
            type === PNMLModelTypes.REF_PLACE ||
            type === PNMLModelTypes.REF_TRANSITION
        ) {
            return true;
        }
        return false;
    }
}

export class ServiceNode extends CategoryNode {
    override isContainableElement(input: GModelElement | GModelElementSchema | string): boolean {
        const type: string = getTypeString(input);
        if (type === PNMLModelTypes.SERVICE_CHANNEL) {
            return true;
        }
        return false;
    }
}

export class PNMLGraph extends GGraph {
    override isContainableElement(_input: string | GShapeElement | GModelElementSchema): boolean {
        const type: string = getTypeString(_input);
        if (type === PNMLModelTypes.PAGE_NODE) {
            return true;
        }
        return false;
    }
}

export function isTransition(element: GModelElement): boolean {
    return element.type === PNMLModelTypes.TRANSITION;
}

export function isPlace(element: GModelElement): boolean {
    return element.type === PNMLModelTypes.PLACE;
}


export function isRefTransition(element: GModelElement): boolean {
    return element.type === PNMLModelTypes.REF_TRANSITION;
}

export function isRefPlace(element:GModelElement): boolean {
    return element.type === PNMLModelTypes.REF_PLACE;
}


export interface HasArgs {
    args: any;
}
