/********************************************************************************
 * Copyright (c) 2024 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import {
    AbstractUIExtension,
    codiconCSSClasses,
    Disposable,
    DOMHelper,
    EditorContextService,
    GLSPActionDispatcher,
    GLSPMouseTool,
    GModelRoot,
    GNode,
    hasArgs,
    ModelIndexImpl,
    MousePositionTracker,
    TYPES,
    ViewerOptions
} from '@eclipse-glsp/client';
import { inject, injectable } from 'inversify';
import { AddDiNeROSTokenOperation } from '../Actions/AddDiNeROSTokenOperation';
import { DeleteDiNeROSTokenOperation } from '../Actions/DeleteDiNeROSTokenOperation';
import { ModelUpdateListener } from '../Handlers/ModelUpdateListener';
import { isPlace } from '../model';

@injectable()
export class EditTokensMenu extends AbstractUIExtension {
    @inject(TYPES.IActionDispatcher)
    protected actionDispatcher: GLSPActionDispatcher;

    @inject(EditorContextService)
    protected editorContextService: EditorContextService;

    @inject(MousePositionTracker)
    protected mousePositionTracker: MousePositionTracker;

    @inject(TYPES.ViewerOptions)
    protected viewerOptions: ViewerOptions;

    @inject(TYPES.DOMHelper)
    protected domHelper: DOMHelper;

    @inject(GLSPMouseTool) protected mouseTool: GLSPMouseTool;

    @inject(ModelUpdateListener)
    protected modelUpdateListener: ModelUpdateListener;

    protected waitForUpdate: Disposable;
    protected noUpdateInProgress = true;

    static readonly ID = 'edit-tokens-menu';

    protected place: GNode;

    override id(): string {
        return EditTokensMenu.ID;
    }
    override containerClass(): string {
        return 'command-palette';
    }
    protected override initializeContents(containerElement: HTMLElement): void {
        containerElement.appendChild(this.createWidget());
    }

    protected createWidget(): HTMLElement {
        const editingElement = document.createElement('div');
        editingElement.id = 'token-editing-element';
        editingElement.classList.add('edit-box');
        const header = document.createElement('div');

        header.appendChild(document.createTextNode('Edit Tokens'));

        const closeIcon = this.createIcon('chrome-close', 'Close');
        closeIcon.style.margin = '0px 0px 0px 0px';

        closeIcon.onclick = () => {
            this.hide();
        };

        header.appendChild(closeIcon);
        header.classList.add('edit-box-header');
        editingElement.appendChild(header);

        const placeIdDiv = document.createElement('div');
        placeIdDiv.appendChild(document.createTextNode('Place ID: ' + this.place.id));
        placeIdDiv.style.padding = '0.4em';
        editingElement.appendChild(placeIdDiv);

        editingElement.appendChild(this.createInputSection());
        const sepLine = document.createElement('hr');
        sepLine.style.margin = '4px 0px 0px';

        editingElement.appendChild(sepLine);
        editingElement.appendChild(this.createDeletionSection());
        return editingElement;
    }

    protected createInputSection(): HTMLElement {
        const inputSection = document.createElement('div');
        const titleDiv = document.createElement('div');
        const title = document.createTextNode('Add Token:');
        titleDiv.appendChild(title);
        titleDiv.classList.add('edit-box-text');

        inputSection.appendChild(titleDiv);
        const inputField = document.createElement('textarea');
        inputField.id = 'add-token-input-field';
        inputField.style.width = '90%';
        inputField.style.border = 'solid';
        inputField.style.borderWidth = '1px';
        inputField.style.borderColor = '#DDDDDD';
        inputSection.appendChild(inputField);

        const addButton = document.createElement('button');
        addButton.appendChild(document.createTextNode('Add'));
        addButton.style.color = '#333333';
        addButton.onclick = async () => {
            if (this.noUpdateInProgress) {
                const token = inputField.value;
                if (!token) {
                    return;
                }
                await this.actionDispatcher.dispatch(AddDiNeROSTokenOperation.create(this.place.id, token));

                this.noUpdateInProgress = false;
            }
        };
        inputSection.appendChild(addButton);
        return inputSection;
    }

    protected createDeletionSection(): HTMLElement {
        const deletionSection = document.createElement('div');
        const titleDiv = document.createElement('div');
        const title = document.createTextNode('Tokens:');
        titleDiv.appendChild(title);
        titleDiv.classList.add('edit-box-text');

        deletionSection.appendChild(titleDiv);

        const placeArgs = hasArgs(this.place) ? this.place.args : {};
        const tokenListString = placeArgs.BalloonTokens;
        const tokens: string[] = JSON.parse(tokenListString.toString());
        tokens.forEach(token => {
            deletionSection.appendChild(this.createDeletionElement(token));
        });

        return deletionSection;
    }

    protected createDeletionElement(token: string): HTMLElement {
        const deletionElement = document.createElement('div');
        deletionElement.classList.add('delete-field');
        deletionElement.style.paddingBottom = '5px';
        const icon = this.createIcon('chrome-close', 'Delete');

        icon.onclick = async () => {
            if (this.noUpdateInProgress) {
                await this.actionDispatcher.dispatch(DeleteDiNeROSTokenOperation.create(this.place.id, token));

                this.noUpdateInProgress = false;
            }
        };
        deletionElement.appendChild(icon);
        deletionElement.insertAdjacentText('beforeend', token);
        return deletionElement;
    }

    protected createIcon(codiconId: string, hoverText: string): HTMLElement {
        const icon = document.createElement('i');
        icon.classList.add(...codiconCSSClasses(codiconId), 'overlay-icon', 'delete-icon');
        icon.setAttribute('title', hoverText);
        return icon;
    }

    override show(root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.place = getPlace(contextElementIds, root.index)[0];
        super.show(root, ...contextElementIds);
        document.getElementById('token-editing-element')?.replaceWith(this.createWidget());
        this.waitForUpdate = this.modelUpdateListener.onModelUpdateComplete(change => {
            this.modelUpdateReload();
        });
    }

    protected override onBeforeShow(containerElement: HTMLElement, root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.setPosition(containerElement);
    }

    protected setPosition(containerElement: HTMLElement): void {
        this.containerElement.style.position = 'absolute';
        const x = 250;
        const y = 100;

        containerElement.style.left = `${x}px`;
        containerElement.style.top = `${y}px`;
        containerElement.style.width = '520px';
    }

    override hide(): void {
        this.waitForUpdate?.dispose();
        super.hide();
    }

    modelUpdateReload(): void {
        this.place = getPlace([this.place.id], this.editorContextService.modelRoot.index)[0];
        document.getElementById('token-editing-element')?.replaceWith(this.createWidget());
        this.noUpdateInProgress = true;
    }
}

function getPlace(ids: string[], index: ModelIndexImpl): GNode[] {
    return ids.map(id => index.getById(id)).filter(element => element && isPlace(element)) as GNode[];
}
