/********************************************************************************
 * Copyright (c) 2024 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/
import {
    angleOfPoint,
    GEdge,
    GNode,
    GShapeElement,
    Hoverable,
    isSelectable,
    IViewArgs,
    Point,
    PolylineEdgeViewWithGapsOnIntersections,
    RenderingContext,
    RoundedCornerNodeView,
    SEdgeImpl,
    Selectable,
    svg,
    toDegrees
} from '@eclipse-glsp/client';
import { injectable } from 'inversify';
import { VNode } from 'snabbdom';
import { HasArgs } from './model';

// eslint-disable-next-line @typescript-eslint/no-unused-vars
const JSX = { createElement: svg };

@injectable()
export class DirectedEdgeView extends PolylineEdgeViewWithGapsOnIntersections {
    protected arrowHead = 'M 1,0 L 10,-4 L 10,4 Z';

    protected override renderAdditionals(edge: GEdge, segments: Point[], context: RenderingContext): VNode[] {
        const additionals = super.renderAdditionals(edge, segments, context);
        const p1 = segments[segments.length - 2];
        const p2 = segments[segments.length - 1];

        // edge.parent.move(edge, edge.parent.children.length - 1);

        const arrow = (
            <path
                class-sprotty-edge={true}
                class-arrow={true}
                d={this.arrowHead}
                transform={`rotate(${toDegrees(angleOfPoint({ x: p1.x - p2.x, y: p1.y - p2.y }))} ${p2.x} ${p2.y}) translate(${p2.x} ${
                    p2.y
                })`}
            />
        );
        additionals.push(arrow);
        return additionals;
    }
}

@injectable()
export class ReadArcView extends DirectedEdgeView {
    protected override arrowHead = 'M 1,0 L 10,-5 L 19,0 L 10,5 Z';
}

@injectable()
export class InhibitorArcView extends DirectedEdgeView {
    protected override arrowHead = 'M 1,0 m 5,0 a 5,5 0 1,0 -10,0 a 5,5 0 1,0 10,0';
}

@injectable()
export class PageNodeView extends RoundedCornerNodeView {
    override render(node: Readonly<GNode & Hoverable & Selectable & HasArgs>, context: RenderingContext): VNode | undefined {
        if (node.type === 'pageNode' && node.args.hidden !== undefined && node.args.hidden === true) {
            return undefined;
        }
        return super.render(node, context);
    }

    protected override renderWithoutRadius(
        node: Readonly<GShapeElement & Hoverable & Selectable & HasArgs>,
        context: RenderingContext
    ): VNode | undefined {
        if (node.type === 'pageNode' && node.args.hidden !== undefined && node.args.hidden === true) {
            return undefined;
        }
        return super.renderWithoutRadius(node, context);
    }
}

@injectable()
export class ReferenceEdgeView extends PolylineEdgeViewWithGapsOnIntersections {
    override render(edge: Readonly<SEdgeImpl>, context: RenderingContext, args?: IViewArgs): VNode | undefined {
        const source = edge.source;
        if (source && isSelectable(source) && source.selected) {
            return super.render(edge, context, args);
        }
        return undefined;
    }
}
