package org.eclipse.glsp.example.javaemf.Simulation;

import com.google.inject.Inject;
import org.eclipse.glsp.example.javaemf.model.ExtendedModelState;
import org.eclipse.glsp.example.javaemf.model.ModelAccessUtil;
import org.eclipse.glsp.example.javaemf.operation.AddDiNeROSTokenOperation;
import org.eclipse.glsp.example.javaemf.operation.DeleteDiNeROSTokenOperation;
import org.eclipse.glsp.example.javaemf.validators.InputSignalValidator;
import org.eclipse.glsp.server.actions.ActionDispatcher;
import org.pnml.emf.ptnet.*;

import java.util.Map;

public class FireTransitionExecutor {
    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected ModelAccessUtil modelAccessUtil;

    @Inject
    protected ActionDispatcher actionDispatcher;

    @Inject
    protected InputSignalValidator inputSignalValidator;

    public boolean isFireable(Transition transition) {
        return isSignalActive(transition) && enoughTokensInPlaces(transition);
    }

    public boolean enoughTokensInPlaces(TransitionNode transition) {
        for(Arc arc : transition.getInArcs()){
            PlaceNode sourceNode = (PlaceNode) arc.getSource();
            Place place = modelAccessUtil.resolveReferencePlace(sourceNode);
            if (place == null){
                // the refPlace has no reference and is therefore ignored
                continue;
            }
            int tokensInPlace = modelAccessUtil.getTokensInPlace(place);
            int arcWeight = modelAccessUtil.getArcWeight(arc);
            if(modelAccessUtil.isInhibitorArc(arc)){
                if(arcWeight <= tokensInPlace){
                    return false;
                }
            } else {
                if(arcWeight > tokensInPlace){
                    return false;
                }
            }

        }

        for (RefTransition refTransition : transition.getReferencingTransitions()){
            if (!enoughTokensInPlaces(refTransition)){
                return false;
            }
        }
        return true;
    }

    public boolean isNotInhibited(Transition transition) {
        for(Arc arc : transition.getInArcs()){
            if(modelAccessUtil.isInhibitorArc(arc)){
                Place place = (Place) arc.getSource();
                int tokensInPlace = modelAccessUtil.getTokensInPlace(place);
                int arcWeight = modelAccessUtil.getArcWeight(arc);

                if(arcWeight <= tokensInPlace){
                    return false;
                }
            }
        }
        return true;
    }

    public boolean isSignalActive(Transition transition) {
        return inputSignalValidator.evaluateCNF(modelAccessUtil.getInputSignalClause(transition));
    }

    // takes a place or a refplace and returns the place or the place the redplace references


    public void fireTransition(TransitionNode transition) {
        if(transition instanceof Transition transition1 && !isFireable(transition1)) {
            return;
        }

        transition.getInArcs().forEach(arc -> {
            if(!modelAccessUtil.isInhibitorArc(arc) && !modelAccessUtil.isReadArc(arc)){
                PlaceNode sourceNode = (PlaceNode) arc.getSource();
                Place place = modelAccessUtil.resolveReferencePlace(sourceNode);
                if (place == null){
                    // the refPlace has no reference and is therefore ignored
                    return; //only returns the lambda
                }

                int tokensInPlace = modelAccessUtil.getTokensInPlace(place);
                int arcWeight = modelAccessUtil.getArcWeight(arc);

                modelState.getSimulationMarking().put(place.getId(), tokensInPlace - arcWeight);
            }
        });

        transition.getOutArcs().forEach(arc -> {
            if(!modelAccessUtil.isInhibitorArc(arc) && !modelAccessUtil.isReadArc(arc)){
                PlaceNode sourceNode = (PlaceNode) arc.getTarget();
                Place place = modelAccessUtil.resolveReferencePlace(sourceNode);
                if (place == null){
                    // the refPlace has no reference and is therefore ignored
                    return; //only returns the lambda
                }

                int tokensInPlace = modelAccessUtil.getTokensInPlace(place);
                int arcWeight = modelAccessUtil.getArcWeight(arc);

                modelState.getSimulationMarking().put(place.getId(), tokensInPlace + arcWeight);
            }
        });
        for (RefTransition refTransition : transition.getReferencingTransitions()) {
            fireTransition(refTransition);
        }
    }


    public void fireTransitionWithBalloonToken(Transition transition, Map<String, String[]> placeTokensMap) {
        if(!isFireable(transition)) {
            return;
        }

        StringBuilder newToken = new StringBuilder();

        gatherAndRemoveTokens(transition, placeTokensMap, newToken);

        placeTokens(transition, newToken);
    }

    protected void gatherAndRemoveTokens(TransitionNode transition, Map<String, String[]> placeTokensMap, StringBuilder newToken){
        for (Arc arc : transition.getInArcs()){
            if(!modelAccessUtil.isInhibitorArc(arc) && !modelAccessUtil.isReadArc(arc)){
                PlaceNode sourceNode = (PlaceNode) arc.getSource();
                Place place = modelAccessUtil.resolveReferencePlace(sourceNode);
                if (place == null){
                    // the refPlace has no reference and is therefore ignored
                    continue;
                }

                String[] tokens = placeTokensMap.get(place.getId());
                if(tokens == null){
                    tokens = new String[0];
                }
                int arcWeight = modelAccessUtil.getArcWeight(arc);
                int i = 0;
                while (i < tokens.length && i < arcWeight){
                    newToken.append(tokens[i]);
                    actionDispatcher.dispatch(new DeleteDiNeROSTokenOperation(place.getId(), tokens[i]));
                    i++;
                }

                while (i < arcWeight){
                    newToken.append(modelAccessUtil.getBalloonTokens(place).get(0));
                    actionDispatcher.dispatch(new DeleteDiNeROSTokenOperation(place.getId(), modelAccessUtil.getBalloonTokens(place).get(0)));
                    i++;
                }
            }
        }
        for (RefTransition refTransition : transition.getReferencingTransitions()){
            // recursively calls this method for all referencing transitions
            gatherAndRemoveTokens(refTransition, placeTokensMap, newToken);
        }
    }

    protected void placeTokens(TransitionNode transition, StringBuilder newToken){
        for (Arc arc : transition.getOutArcs()){
            if(!modelAccessUtil.isInhibitorArc(arc) && !modelAccessUtil.isReadArc(arc)){
                PlaceNode sourceNode = (PlaceNode) arc.getTarget();
                Place place = modelAccessUtil.resolveReferencePlace(sourceNode);
                if (place == null){
                    // the refPlace has no reference and is therefore ignored
                    continue;
                }

                int arcWeight = modelAccessUtil.getArcWeight(arc);

                for (int i = 0; i < arcWeight; i++){
                    actionDispatcher.dispatch(new AddDiNeROSTokenOperation(place.getId(), newToken.toString()));
                }
            }
        }
        for (RefTransition refTransition : transition.getReferencingTransitions()){
            // recursively calls this method for all referencing transitions
            placeTokens(refTransition, newToken);
        }
    }



}
