package org.eclipse.glsp.example.javaemf.TinaMenu;

import com.google.inject.Inject;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.glsp.example.javaemf.action.RunTinaAction;
import org.eclipse.glsp.example.javaemf.action.TinaResponseAction;
import org.eclipse.glsp.example.javaemf.model.ExtendedModelState;
import org.eclipse.glsp.server.actions.AbstractActionHandler;
import org.eclipse.glsp.server.actions.Action;
import org.pnml.emf.ptnet.Arc;
import org.pnml.emf.ptnet.pnml;

import java.io.*;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;

public class RunTinaActionHandler extends AbstractActionHandler<RunTinaAction> {

    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected PNMLFlattener flattener;

    boolean isWindows;

    HashSet<String> supportedTools = new HashSet<>(Arrays.asList(
        "tina",
        "sift",
        "reduce",
        "pathto",
        "ndrio"
    ));

    public RunTinaActionHandler() {
        isWindows = System.getProperty("os.name").toLowerCase().startsWith("windows");
    }


    @Override
    protected List<Action> executeAction(RunTinaAction runTinaAction) {
        String[] tinaArgs = runTinaAction.getTinaArgs().split(" ");

        String usedTool = tinaArgs[0];

        if(!supportedTools.contains(usedTool)){
            return listOf(new TinaResponseAction(usedTool + " is not a supported tool."));
        }

        ArrayList<String> tinaArgsList = Arrays.stream(tinaArgs)
            .collect(Collectors.toCollection(ArrayList::new));
        tinaArgsList.remove(0);


        flattener.flatten();

        URI uri = modelState.getSemanticModel().eResource().getURI();

        //String homeDir = System.getProperty("user.dir");
        //String tinaLocation = homeDir + "/libs/tina-3.8.0/bin";

        URL location = RunTinaActionHandler.class.getProtectionDomain().getCodeSource().getLocation();
        String tinaLocation = location.getFile()  + "/../../../libs/tina-3.8.0/bin";


        ProcessBuilder processBuilder = new ProcessBuilder();

        processBuilder.directory(new File(tinaLocation));

        URI dirUri = uri.trimSegments(1);

        URI outputDirUri = dirUri.appendSegment("tinaOutput");

        File dir = new File(outputDirUri.device() + outputDirUri.path());
        if (!dir.exists()) {
            dir.mkdirs();
        }

        URI flatPNMLUri = outputDirUri.appendSegment("flat.pnml");
        String flatPNMLPath = flatPNMLUri.devicePath();
        URI netFileUri = outputDirUri.appendSegment("flat.net");
        String netFilePath = netFileUri.devicePath();
        URI netFileWithInhibitorsUri = outputDirUri.appendSegment("inhibitedFlat.net");
        String netFileWithInhibitorsPath = netFileWithInhibitorsUri.devicePath();

        createFlattenedPNMLFile(flatPNMLUri);
        createNetFile(tinaLocation, flatPNMLPath, netFilePath);
        addInhibitorArcsToNetFile(netFilePath, netFileWithInhibitorsPath);



        String outputFileName = switch (usedTool) {
            case "sift" ->"Output.ktz";
            case "reduce" -> "Output.net";
            case "ndrio" -> "Output";
            default -> "Output.txt";
        };

        String outputPath =  outputDirUri.device() + outputDirUri.appendSegment(usedTool + outputFileName).path();

        List<String> command = new ArrayList<>();

        command.add(tinaLocation + "/" + usedTool);
        command.addAll(tinaArgsList);

        if(usedTool.equals("pathto")){
            URI ktzioFileUri = outputDirUri.appendSegment("siftFile.ktz");
            String ktzioFilePath = ktzioFileUri.devicePath();
            createKtzFile(tinaLocation, netFileWithInhibitorsPath, ktzioFilePath);
            command.add(ktzioFilePath);
        } else {
            command.add(netFileWithInhibitorsPath);
        }
        command.add(outputPath);

        processBuilder.command(command);


        String response = runProcess(processBuilder, outputPath);

        if(usedTool.equals("sift")){
            URI txtFileUri = outputDirUri.appendSegment(usedTool + "Output.txt");
            String txtFileFilePath = txtFileUri.devicePath();
            convertToTxt(tinaLocation, outputPath, txtFileFilePath);
        }

        return listOf(new TinaResponseAction(response));
    }



    protected String runProcess(ProcessBuilder processBuilder, String outputPath) {
        try {
            File file = new File(outputPath);
            file.createNewFile();

            Process process = processBuilder.start();

            BufferedReader errorReader = new BufferedReader(
                new InputStreamReader(process.getErrorStream()));

            StringBuilder errorOutput = new StringBuilder();
            String errorLine;
            while ((errorLine = errorReader.readLine()) != null) {
                errorOutput.append(errorLine).append("\n");
            }

            int exitVal = process.waitFor();
            if (exitVal == 0) {
                System.out.println("Success\n");
                return "Success";
            } else {
                System.out.println("Failure\n");
                System.out.println(errorOutput);
                return "Failure: \n" + errorOutput;
            }

        } catch (IOException | InterruptedException e) {
            e.printStackTrace();
        }

        return "Failure";

    }

    protected void createFlattenedPNMLFile(URI flatPNMLUri){
        pnml flatPNML = flattener.flatten();

        XMIResource newResource = new XMIResourceImpl(flatPNMLUri);
        newResource.setEncoding("UTF-8");
        newResource.getContents().add(flatPNML);
        try {
            newResource.save(Collections.EMPTY_MAP);
        } catch (IOException e) {
            System.out.println("failure");
            e.printStackTrace();
        }
    }


    protected void createNetFile(String tinaLocation, String inputPath, String outputPath){
        ProcessBuilder ndrioProcessBuilder = new ProcessBuilder();
        ndrioProcessBuilder.command(tinaLocation + "/ndrio", "-PNML", inputPath, outputPath);

        try {
            Process process = ndrioProcessBuilder.start();

            int exitVal = process.waitFor();
            if (exitVal == 0) {
                return;
            }

        } catch (IOException | InterruptedException e) {
            e.printStackTrace();
        }

    }

    protected void createKtzFile(String tinaLocation, String inputPath, String outputPath){
        ProcessBuilder ndrioProcessBuilder = new ProcessBuilder();
        ndrioProcessBuilder.command(tinaLocation + "/sift", inputPath, outputPath);

        try {
            Process process = ndrioProcessBuilder.start();

            BufferedReader errorReader = new BufferedReader(
                new InputStreamReader(process.getErrorStream()));
            StringBuilder errorOutput = new StringBuilder();
            String errorLine;
            while ((errorLine = errorReader.readLine()) != null) {
                errorOutput.append(errorLine).append("\n");
            }

            int exitVal = process.waitFor();
            if (exitVal == 0) {
                System.out.println("KTZ Success");
            } else {
                System.out.println("KTZ Failure");
                System.out.println(errorOutput);
            }

        } catch (IOException | InterruptedException e) {
            e.printStackTrace();
        }

    }

    protected void addInhibitorArcsToNetFile(String netFilePath, String netFileWithInhibitorsPath){
        final String NET_INHIBITOR_ARC = "?-1";
        ArrayList<Arc> inhibitorArcs = flattener.getInhibitorArcs();
        StringBuilder finalContent = new StringBuilder();

        try (BufferedReader br = new BufferedReader(new FileReader(netFilePath))) {
            String line;
            while ((line = br.readLine()) != null) {
                // process line by line
                if(line.startsWith("tr")){

                    String[] splitLine = line.split(" ");
                    boolean matched = false;

                    for(Arc a : inhibitorArcs){
                        if(a.getTarget().getId().equals(splitLine[1].
                            replace("{", "").
                            replace("}", "").
                            replace(" ", ""))){

                            StringBuilder finalLineContent = new StringBuilder();
                            for(int i = 0; i < splitLine.length; i++){

                                if(a.getSource().getId().equals(splitLine[i].
                                    replace("{", "").
                                    replace("}", "").
                                    replace(" ", ""))){

                                    finalLineContent.append(" ");
                                    finalLineContent.append(splitLine[i]);
                                    finalLineContent.append(NET_INHIBITOR_ARC);

                                } else {
                                    if ( i != 0){
                                        finalLineContent.append(" ");
                                    }
                                    finalLineContent.append(splitLine[i]);
                                }
                            }

                            finalContent.append(finalLineContent);
                            finalContent.append("\n");
                            matched = true;
                            break;
                        }
                    }

                    if(!matched) {
                        finalContent.append(line);
                        finalContent.append("\n");
                    }
                }else {
                    finalContent.append(line);
                    finalContent.append("\n");
                }
            }

            BufferedWriter writer = new BufferedWriter(new FileWriter(new File(netFileWithInhibitorsPath)));
            writer.write(finalContent.toString());
            writer.flush();
            writer.close();

        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    protected void convertToTxt(String tinaLocation, String inputPath, String outputPath){
        ProcessBuilder ndrioProcessBuilder = new ProcessBuilder();

        ndrioProcessBuilder.command(tinaLocation + "/ktzio", inputPath, outputPath);

        try {
            Process process = ndrioProcessBuilder.start();

            BufferedReader errorReader = new BufferedReader(
                new InputStreamReader(process.getErrorStream()));
            StringBuilder errorOutput = new StringBuilder();
            String errorLine;
            while ((errorLine = errorReader.readLine()) != null) {
                errorOutput.append(errorLine).append("\n");
            }

            int exitVal = process.waitFor();
            if (exitVal == 0) {
                System.out.println("KTZ Success");
            } else {
                System.out.println("KTZ Failure");
                System.out.println(errorOutput);
            }
        } catch (IOException | InterruptedException e) {
            e.printStackTrace();
        }

    }

}


