package org.eclipse.glsp.example.javaemf.model;

import jakarta.inject.Inject;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.graph.GPoint;
import org.eclipse.glsp.graph.impl.GPointImpl;
import org.pnml.emf.ptnet.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

// utility class for model access to avoid null checks in accessing classes
public class ModelAccessUtil {
    @Inject
    ExtendedModelState modelState;

    public int getTokensInPlace(Place place){
        int tokensInPlace;
        if(modelState.getSimulationMarking().get(place.getId()) != null){
            tokensInPlace = modelState.getSimulationMarking().get(place.getId());
        } else if(place.getInitialMarking() != null) {
            tokensInPlace = place.getInitialMarking().getText();
        } else {
            tokensInPlace = 0;
        }
        return tokensInPlace;
    }

    public int getInitialMarking(Place place){
        if (place.getInitialMarking() != null){
            return place.getInitialMarking().getText();
        } else {
            return 0;
        }
    }

    public int getArcWeight(Arc arc){
        int arcWeight;
        if(arc.getInscription() != null){
            arcWeight = arc.getInscription().getText();
        } else {
            arcWeight = 1;
        }
        return arcWeight;
    }

    public GPoint getPosition (PositionableObject element){
        GPoint position = new GPointImpl();
        position.setX(Optional.ofNullable(element.getNodegraphics()).map(NodeGraphics::getPosition).map(Coordinate::getX).orElse(0));
        position.setY(Optional.ofNullable(element.getNodegraphics()).map(NodeGraphics::getPosition).map(Coordinate::getY).orElse(0));
        return position;
    }

    public String getName (PnObject element){
        if(element.getName() == null){
            return "untitled";
        }
        return  element.getName().getText();
    }

    public Place resolveReferencePlace (PlaceNode placeNode){
        if (placeNode instanceof Place place){
            return place;
        }

        while (placeNode instanceof RefPlace refPlace){
            placeNode = refPlace.getRef();
        }
        if (placeNode instanceof Place place){
            return place;
        }
        return null;
    }

    public Transition resolveReferenceTransition (TransitionNode transitionNode){
        if (transitionNode instanceof Transition transition){
            return transition;
        }

        while (transitionNode instanceof RefTransition refTransition){
            transitionNode = refTransition.getRef();
        }
        if (transitionNode instanceof Transition transition){
            return transition;
        }
        return null;
    }

    public boolean isTopicTransition(Transition transition){
        if(transition.getToolspecific() == null || transition.getToolspecific().isEmpty()){
            return false;
        }
        return transition.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.TOPIC_TRANSITION));
    }

    public boolean isServiceTransition(Transition transition) {
        if(transition.getToolspecific() == null || transition.getToolspecific().isEmpty()){
            return false;
        }
        return transition.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.SERVICE_TRANSITION));
    }

    public boolean isDiscreteTransition(Transition transition) {
        if(transition.getToolspecific() == null || transition.getToolspecific().isEmpty()){
            return true;
        }
        return transition.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.DISCRETE_TRANSITION))
            || transition.getToolspecific().stream()
            .noneMatch(toolInfo -> Objects.equals(toolInfo.getTool(), PNMLModelTypes.DINEROS_TOOL_NAME));
    }

    public boolean isNodePageNode(Page page)
    {
        if(page.getToolspecific() == null || page.getToolspecific().isEmpty()){
            return false;
        }
        return page.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.NODE_PAGE_NODE));
    }

    public boolean isServerPage(Page page) {
        if(page.getToolspecific() == null || page.getToolspecific().isEmpty()){
            return false;
        }
        return page.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.SERVER_PAGE));
    }

    public Optional<ToolInfoTransition> getDiNeROSTransitionToolSpecific(Transition transition, String type){
        return transition.getToolspecific().stream().filter(toolInfo -> Objects.equals(toolInfo.getType(), type))
            .findFirst();
    }

    public Optional<ToolInfoNet> getDiNeROSNetToolSpecific(PetriNet net){
        return net.getToolspecific().stream().filter(toolInfo -> Objects.equals(toolInfo.getTool(), PNMLModelTypes.DINEROS_TOOL_NAME))
            .findFirst();
    }

    public Optional<ToolInfoPage> getDiNeROSPageToolSpecific(Page page){
        return page.getToolspecific().stream().filter(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.SERVER_PAGE))
            .findFirst();
    }

    public String getInputSignalClause(Transition transition){
        return transition.getToolspecific().stream().filter(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.DISCRETE_TRANSITION))
            .findFirst().map(ToolInfoTransition::getInputsignalclause).orElse("");
    }
    /*
    public Map<String, Boolean> getInputSignals(){
        pnml netContainer = modelState.getSemanticModel(pnml.class).orElseThrow();
        PetriNet petriNet = netContainer.getNet().stream().findFirst().orElseThrow();
        return getDiNeROSNetToolSpecific(petriNet).map(ToolInfoNet::getInputsignals)
            .map(InputSignals::getInputsignal).map(inputSignals ->
                inputSignals.stream().filter(inputSignal -> Objects.equals(inputSignal.getInputsignalID(), literal))
                    .anyMatch(InputSignal::isInitialvalue)
            ).orElse(false);
    }
*/

    public Optional<ToolInfoPlace> getDiNeROSPlaceToolSpecific(Place place){
        return place.getToolspecific().stream().filter(toolInfo -> Objects.equals(toolInfo.getTool(), PNMLModelTypes.DINEROS_TOOL_NAME))
            .findFirst();
    }

    public List<String> getBalloonTokens(Place place){
        Optional<List<String>> optTokens = getDiNeROSPlaceToolSpecific(place)
            .map(ToolInfoPlace::getBalloonMarking)
            .map(BalloonMarking::getTokens)
            .map(Tokens::getToken);

        return optTokens.orElseGet(ArrayList::new);

    }

    public boolean isInhibitorArc(Arc arc){
        if(arc.getToolspecific() == null || arc.getToolspecific().isEmpty()){
            return false;
        }
        return arc.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.INHIBITOR_ARC_TYPE));
    }

    public boolean isReadArc(Arc arc){
        if(arc.getToolspecific() == null || arc.getToolspecific().isEmpty()){
            return false;
        }
        return arc.getToolspecific().stream()
            .anyMatch(toolInfo -> Objects.equals(toolInfo.getType(), PNMLModelTypes.READ_ARC_TYPE));
    }

}
