/********************************************************************************
 * Copyright (c) 2022 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * https://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied:
 * -- GNU General Public License, version 2 with the GNU Classpath Exception
 * which is available at https://www.gnu.org/software/classpath/license.html
 * -- MIT License which is available at https://opensource.org/license/mit.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0 OR MIT
 ********************************************************************************/
package org.eclipse.glsp.example.javaemf.model;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.ResourceSet;
//import org.eclipse.glsp.example.tasklist.model.ModelPackage;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.glsp.example.javaemf.MQTT.MQTTClient;
import org.eclipse.glsp.server.actions.SaveModelAction;
import org.eclipse.glsp.server.emf.EMFSourceModelStorage;
import org.eclipse.glsp.server.emf.notation.EMFNotationSourceModelStorage;
import org.eclipse.glsp.server.features.core.model.RequestModelAction;
import org.eclipse.glsp.server.types.GLSPServerException;
import org.eclipse.glsp.server.utils.ClientOptionsUtil;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.pnml.emf.ptnet.*;

import org.eclipse.emf.ecore.resource.Resource;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;

/* see https://github.com/eclipse-glsp/glsp-server/blob/master/plugins/org.eclipse.glsp.server.emf/src/org/eclipse/glsp/server/emf/EMFSourceModelStorage.java
       https://github.com/eclipse-glsp/glsp-server/blob/master/plugins/org.eclipse.glsp.server.emf/src/org/eclipse/glsp/server/emf/notation/EMFNotationSourceModelStorage.java
       https://download.eclipse.org/modeling/emf/emf/javadoc/2.11/org/eclipse/emf/common/util/URI.html
 */

public class TaskListSourceModelStorage extends EMFNotationSourceModelStorage {


   protected static Logger LOGGER = LogManager.getLogger(EMFSourceModelStorage.class.getSimpleName());

   @Override
   protected ResourceSet setupResourceSet(final ResourceSet resourceSet) {
      resourceSet.getPackageRegistry().put(PtnetPackage.eINSTANCE.getNsURI(), PtnetPackage.eINSTANCE);
      return super.setupResourceSet(resourceSet);
   }

   @Override
   protected void doLoadSourceModel(final ResourceSet resourceSet, final URI sourceURI,
                                    final RequestModelAction action) {
      loadSemanticModel(resourceSet, sourceURI, action);
   }

   /*
    * taken directly from the EMFSourceModelStorage class
    *    
    * check here if the given uri points to an empty file; if so load model from the base.pnml file instead
    *    if the model is saved next, the changes will be written to the epty file
    *
    *    this works, because the saving location is provided by the action object (message to the server),
    *    not in the source model
    */


   @Override
   public void loadSourceModel(final RequestModelAction action) {
      String sourceURI = ClientOptionsUtil.getSourceUri(action.getOptions())
         .orElseThrow(() -> new GLSPServerException("No source URI given to load model!"));
      URI resourceURI = URI.createFileURI(sourceURI);

      ResourceSet resourceSet = getOrCreateEditingDomain().getResourceSet();

      try{
         // check how many bytes the inputstream from the url yields before hitting some interrupt
         if(0 == resourceSet.getURIConverter().createInputStream(resourceURI).available())
            setupAndLoadNewModel(resourceSet, resourceURI, action);

         else
            doLoadSourceModel(resourceSet, resourceURI, action);
      }
      catch(IOException e)
      {
         System.err.println("\nIO Exception trying to access uri: "+ resourceURI +"\n");
      }
   }


   private void setupAndLoadNewModel(final ResourceSet resourceSet, final URI sourceURI,
      final RequestModelAction action)
   {
      /*
       * this function creates a local file from which to load a default model, 
       *  then deletes that file after loading is finished
       * 
       * the added model will not be saved and the file will be gone, so there will in fact not be a file source
       * 
       * to actually create the file save it once
       * 
       * possibly the same could be achived by manually building a graphical model and translating it into an empty sorce model
       *  this would avoid the need for additional files
       */
      String fileName = sourceURI.lastSegment();
      int dotIndex = fileName.lastIndexOf(".");
      if(dotIndex >= 0)
         fileName = fileName.substring(0,dotIndex);

      String baseModel = 
         "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n" + //
         "<pnml xmi:version=\"2.0\" xmlns:xmi=\"http://www.omg.org/XMI\" xmlns=\"http://www.pnml.org/version-2009/grammar/pnml\">\r\n" + //
         "  <net id=\""+ fileName +"\"/>\r\n" + //
         //"  </net>\r\n" + //
         "</pnml>";


      // create a server local tempfile for the model to load from
      URI modelURI = null;
      File modelFile = null;
      try{
         modelFile = new File("model.tmp");
         FileWriter modelInput = new FileWriter(modelFile);
         modelInput.write(baseModel);
         modelInput.close();

         modelURI = URI.createFileURI(modelFile.getAbsolutePath());
      }
      catch(IOException e)
      {
         LOGGER.error("Could not create new resource: " + modelURI, e);
         throw new GLSPServerException("Could not save model to file: " + modelURI, e);
      }

      // load the file and set the uri to the intended source
      doLoadSourceModel(resourceSet, modelURI, action);
      Resource resource = resourceSet.getResource(modelURI, false);
      resource.setURI(sourceURI);
      // cleanup temp file
      modelFile.delete();
   }

   /*
    * taken directly from the EMFSourceModelStorage class
    *
    *    to fix export issue, this is problably not the right place -> look into the translator from source model to .xml
    *       potentially the ressource.save has to be extended
    */
   @Override
   public void saveSourceModel(final SaveModelAction action) {
      for (Resource resource : modelState.getResourceSet().getResources()) {
         if (resource.getURI() != null) {
            try {
               resource.save(Collections.EMPTY_MAP);
            } catch (IOException e) {
               LOGGER.error("Could not save resource: " + resource.getURI(), e);
               throw new GLSPServerException("Could not save model to file: " + resource.getURI(), e);
            }
         }
      }
   }
}
