package org.eclipse.glsp.example.javaemf.validators;


import com.google.inject.Inject;
import org.eclipse.glsp.example.javaemf.model.ExtendedModelState;
import org.eclipse.glsp.example.javaemf.model.ModelAccessUtil;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class InputSignalValidator {
    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected ModelAccessUtil modelAccessUtil;

    // the redex to check whether a string is the conjunctive normal form
    // it uses AND, OR and NOT as Boolean symbols
    // it consists of four parts separated by the "|" delimiter
    // the first part checks for "full" CNF formulas containing conjunctions(AND) and bracketed disjunctions(OR)
    //   e.g. "(a OR b) AND (c OR NOT d)", "(a OR b)"
    // the second part checks for formulas without disjunctions
    //   e.g. "a AND b", "a AND b AND c", "(a AND b)" // if the bracketed form should not be valid the optional brackets("\(?") in the beginning and end of the second part can be removed
    // the third part checks for formulas without conjunctions that are not in brackets
    //   e.g. "a OR b", "a OR NOT b OR c"
    // the fourth part checks for an empty string; can be either interpreted as not set or as true by default
    // terms that contain conjunctions and disjunctions and are fully enclosed in brackets are not valid
    //   e.g. "((a OR b) AND (c OR d))" if they should be valid, additional optional brackets ("\(?") can be added around the first part
    // leading and trailing spaces are not allowed
    // not case-sensitive
    // original regex without the additional java escape characters in CNFregex.txt file
    String CNFRegex = "^(\\((NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+(( OR )(NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+)*\\)(( AND )\\((NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+(( OR )(NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+)*\\))*)$|^\\(?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+(( AND )(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+)*\\)?$|^(\\(?(NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+(( OR )(NOT )?(?:(?!\\bAND\\b)(?!\\bOR\\b)(?!\\bNOT\\b)\\w)+)*\\)?)$|^$";
    Pattern pattern = Pattern.compile(CNFRegex, Pattern.CASE_INSENSITIVE);

    public boolean isValidCNF(String input) {
        Matcher matcher = pattern.matcher(input);
        return matcher.find();
    }

    public boolean evaluateCNF(String cnf) {
        if(cnf.isEmpty()){
            // empty expression is true by default
            return true;
        }
        // splits CNF into its separate parts; (?i) for case insensitivity
        String[] disjunctions = cnf.split("(?i) AND ");
        for(String disjunction : disjunctions){
            String[] literals = disjunction.replace("(", "").replace(")", "").split("(?i) OR ");
            boolean disjunctionValue = false;
            for(String literal : literals){
                if(literal.matches("(?i)NOT .*")){
                    if(!getLiteralValue(literal.replaceAll("(?i)NOT ", ""))){
                        disjunctionValue = true;
                        break; // literal value is true therefore disjunction is true therefore rest of disjunction does not need to be evaluated
                    }
                }else {
                    if(getLiteralValue(literal)){
                        disjunctionValue = true;
                        break; // literal value is true therefore disjunction is true therefore rest of disjunction does not need to be evaluated
                    }
                }
            }
            if(!disjunctionValue){
                return false; // no literal was true therefore the disjunction evaluates to false and the CNF also evaluates to false
            }
        }
        return true; // the return false was never reached because all disjunctions evaluated to true
    }

    protected boolean evaluateDisjunction(String disjunction){
        String[] literals = disjunction.replace("(", "").replace(")", "").split("(?i) OR ");
        for(String literal : literals){
            if(literal.matches("(?i)NOT ")){
                if(!getLiteralValue(literal.replaceAll("(?i)NOT ", ""))){
                    return true; // literal value is true therefore disjunction is true therefore rest of disjunction does not need to be evaluated
                }
            }else {
                if(getLiteralValue(literal)){
                    return true; // literal value is true therefore disjunction is true therefore rest of disjunction does not need to be evaluated
                }
            }
        }
        return false; // no literal was true therefore the disjunction evaluates to false and the CNF also evaluates to false
    }

    protected boolean getLiteralValue(String literal){
        Boolean value = modelState.getInputSignalValues().get(literal);
        if(value != null){
            return value;
        }

        // since initial values are supposed to be removed in future DiNeRos versions and are usually false anyway,
        // it is sufficient to just return false otherwise the initial values can be checked with the following code
        /*
        pnml netContainer = modelState.getSemanticModel(pnml.class).orElseThrow();
        PetriNet petriNet = netContainer.getNet().stream().findFirst().orElseThrow();
        return modelAccessUtil.getDiNeROSNetToolSpecific(petriNet).map(ToolInfoNet::getInputsignals)
            .map(InputSignals::getInputsignal).map(inputSignals ->
                inputSignals.stream().filter(inputSignal -> Objects.equals(inputSignal.getInputsignalID(), literal))
                    .anyMatch(InputSignal::isInitialvalue)
            ).orElse(false);
        */

        return false;
    }

}
